#!/usr/bin/env python
# -*- coding: ISO-8859-1 -*-

#------------------
# ScummSpeaks
# Monster.sou class
#------------------
# October 2008
#------------------
# I have no idea when to flush()
#------------------

import os.path
import copy
import cPickle
from os import tmpfile
from shutil import copyfileobj
from ssmisc import *
import ssconst
import ssglob

class MonsterSou(object):
    lDefaultTagList = [0x0F, 0xFF] # default tags
    monTypes = set(["VOC", "MP3", "OGG", "FLAC"])
    
    def __init__(self):
        self.tmpdata = None
        self.reset()
    
    def reset(self):
        if self.tmpdata != None:
            self.tmpdata.flush()
            self.tmpdata.close()
            self.tmpdata = None # tmpfile() created when making new or loading
        self.filename = None # not used
        self.indextable = [] # VOC: just offsets. other: entries of 4 (see loading method)
        self.fakepos = 0 # used for compressed monsters
        self.vocmap = {} # Maps offsets to indices for easier reading
        self.descriptions = []
        self.type = None
    
    def newMonster(self, type):
        """ Accepts a string describing the type. Must be one listed in the static MonsterSou.monTypes property."""
        if type not in MonsterSou.monTypes:
            raise SSException("Unrecognized monster.sou type requested (" + str(type) + ")")
        self.reset()
        self.type = type
        self.tmpdata = tmpfile()
        if self.type is "VOC": # write SOU header for VOC monsters
            self.tmpdata.write(ssconst.strSOU)
    
    def initDescriptions(self):
        self.descriptions = [''] * self.getLength()
    
    def close(self):
        """ Call this when finished with your Monster.
        
        At the moment it just closes the temporary data file (if it exists)."""
        if self.tmpdata != None:
            self.tmpdata.close()
    
    def isClosed(self):
        if self.tmpdata == None or self.tmpdata.closed:
            return True
        return False
    #______________
    # Getters/Setters
    def getLength(self):
        """ Returns the number of sounds in the monster file."""
        # This should be replaced by the more Pythonic use of len()
        return len(self.indextable)
    
    def getIndexTableSize(self):
        """ Returns the size of the index table suitable for saving to a file (ie. in bytes).
        
        Each entry in the table is comprised of four values, each value is a word (4 bytes), so a table
        with one entry will return 16."""
        return len(self.indextable) * 16
    
    def getOriginalOffset(self, i):
        """ Returns the original offset in the monster.sou file (what SCUMM looks up)."""
        #if self.type is "VOC":
        #    return self.indextable[i]
        #else:
        return self.indextable[i][0]
    
    def getActualOffset(self, i):
        """ Returns the actual offset in the monster.sou file. This will be the same as the
        original offset for VOC monsters, compressed monsters will be different."""
        if self.type is "VOC":
            return self.indextable[i][0] # actual offset is the original offset
        else:
            return self.indextable[i][1]
    
    def getDescription(self, i):
        # Used by list display
        if i < len(self.descriptions):
            return self.descriptions[i]
        else:
            return "[no description]"
    
    def getDescriptionFromAddress(self, a):
        d = self.descriptions[self.vocmap[a]]
        if d == "":
            d = self.getDummyDescriptionFromAddress(a)
        return d

    def getDummyDescriptionFromAddress(self, a):
        return str(self.vocmap[a]).zfill(4)

    def getNumTags(self, i):
        # Lip-synch tags not yet implemented from VOC monster.sou files
        if self.type is "VOC":
            return len(self.indextable[i][1])
        else:
            return self.indextable[i][2]
            
        
    def setDescription(self, index, desc):
        """ Accepts an index (logical position of sound in Monster) and the text description"""
        self.descriptions[index] = desc

    def containsAddress(self, a):
        return a in self.vocmap

    # End Getters/Setters
    #______________
    # Save functions
    # (should these be open files or filenames?)
    def saveToFile(self, ofile):
        self.tmpdata.flush()
        self.tmpdata.seek(0)
        if self.type is "VOC":
            self._saveVOCMonster(ofile)
        else:
            self._saveCompressedMonster(ofile)
        
    def _saveVOCMonster(self, ofile):
        copyFileContents(self.tmpdata, ofile)
        
    def _saveCompressedMonster(self, ofile):
        # Write the size of the index table    
        iTableSize = self.getIndexTableSize()
        ofile.write(numToString(iTableSize, 4, True))
        
        # Write the index table
        for entry in self.indextable:
            for item in entry: # entry must only have 4 items, otherwise this will stuff everything up
                ofile.write(numToString(item, 4, True))
      
        # Write the sound data
        copyFileContents(self.tmpdata, ofile)
        
    def saveDescriptions(self, mdfile):
        cPickle.dump(self.descriptions, mdfile)
        
    # End Save functions
    #______________
    # Load functions
    def loadFromFile(self, ofile):
        if os.path.splitext(ofile.name)[1].lower() == ".sou": # real LucasArts monster.sou file, just copy without interpreting
            self.type = "VOC"
            self._loadVOCMonster(ofile)
        else: # Assume it's a ScummVM compressed monster file.
            self._loadCompressedMonster(ofile)
        
    def _loadVOCMonster(self, ofile):
        # Load metadata by analysing the monster.sou file
        self._createVOCTable(ofile)
        
        # Copy monster.sou to temp file
        self.tmpdata = tmpfile()
        ofile.seek(0)
        self.tmpdata.seek(0)
        copyFileContents(ofile, self.tmpdata)
        
    def _createVOCTable(self, ofile):
        # Expects a LucasArts monster.sou file
        ofile.seek(0)
        self.indextable = [] # this stores the offsets and raw lip-sync tags
        self.vocmap = {}
        
        # Go through and record the offsets of each VOC file (actually the
        # VCTL block, as this is what SCUMM looks for when playing sounds).
        while True:
            try:
                header = ofile.read(4)
            except EOFError:
                # Reached end of MONSTER.SOU tempfile
                break
    
            if header == "":
                # Reached end of MONSTER.SOU tempfile
                break
            
            elif header == "SOU ": # SOU is the base block of monster.sou
                ofile.seek(4, 1) # skip the chunk size info
                continue
            
            elif header == "VCTL": # Could possibly add VTTL as it seems to be the same thing
                offset = ofile.tell() - 4
                size = strToNumber(reversed(ofile.read(4)))
                # Store the original offset and the raw tag data
                self.indextable.append( (offset, ofile.read(size - 8)) )
                continue
                
            elif header == "Crea": # skip Crea (this is the actual VOC)
                ofile.seek(22, 1) # skip rest of header (assume it's all fine) - block starts at 0x1A
                while ord(ofile.read(1)) != 0: # while blocktype != 0
                    size = strToNumber(ofile.read(3)) # Read the block size (3 bytes)
                    ofile.seek(size, 1) # skip the block
                continue

        # Set up a reverse map for easier identification. (maps address to index in indextable)
        for i, offset in enumerate(self.indextable):
            self.vocmap[offset[0]] = i
    
    def _loadCompressedMonster(self, ofile):
        self.indextable = []
        self.vocmap = {}
        
        ext = os.path.splitext(ofile.name)[1].lower()
        if ext == ".so3":
            self.type = "MP3"
        elif ext == ".sog":
            self.type = "OGG"
        elif ext == ".sof":
            self.type = "FLAC"
        else:
            raise SSException("Unrecognised monster file type (" + ext + ")")

        i_highest_fake_pos = 0    
        
        i_index_size = strToNumber(ofile.read(4), True)
        
        while ofile.tell() < i_index_size + 4:
            origoff = strToNumber(ofile.read(4), True) # original offset that the game looks up.
            newoff = strToNumber(ofile.read(4), True)  # actual offset in the compressed monster file
            numtags = strToNumber(ofile.read(4), True) # tags for lip-synching (not used by MI2)
            oggsize = strToNumber(ofile.read(4), True) # size of the individual compressed voice file
            self.indextable.append([origoff,
                                newoff,
                                numtags,
                                oggsize])
            if origoff > i_highest_fake_pos: # so we don't collide when adding new sounds
                i_highest_fake_pos = origoff

        # Note that we don't seek to the start, as we only want the tag & OGG data, not
        #  the metadata, in the temp file.
        self.tmpdata = tmpfile()
        copyFileContents(ofile, self.tmpdata)
        self.tmpdata.flush()
        
        # Counter for the highest fake position. Used when adding new sounds.
        self.fakepos = i_highest_fake_pos
        
        # Set up a reverse map for easier identification. (maps address to index in indextable)
        for i, entry in enumerate(self.indextable):
            self.vocmap[entry[0]] = i

    def loadDescriptions(self, mdfile):
        self.descriptions = cPickle.load(mdfile)
        if len(self.descriptions) < len(self.indextable): # this might happen if a VOC has been added. Or not.
            self.descriptions.extend([''] * (len(self.indextable) - len(self.descriptions))) # fill up the rest with empty descriptions
    # End Load functions
    #______________
    # Adding sounds
    def addSound(self, infile, tags=None):
        self.tmpdata.flush()
        self.tmpdata.seek(0)
        
        if tags is None:
            tags = MonsterSou.lDefaultTagList
        
        if self.type is "VOC":
            self._addVOCSound(infile, tags)
        else:
            self._addCompressedSound(infile, tags)
    
    def _addVOCSound(self, vocfile, tags):
        """ Takes an open VOC file object. Checks that it's a valid VOC, then adds it
        to the temporary Monster file. Returns True on success, raises SSExceptions on
        unrecognized data in the VOC file."""
        # Append to the srcfile
        self.tmpdata.seek(0, 2) 
        
        start = self.tmpdata.tell()
        
        # Check that it's actually a recognized VOC file.
        if vocfile.read(19) != 'Creative Voice File':
            raise SSException(vocfile.name + " does not appear to be a VOC file.")
        vocfile.seek(3, 1)
        version1 = strToNumber(vocfile.read(2)) # version is minor/major so this is actually incorrect
        version2 = ~strToNumber(vocfile.read(2)) + 0x1234
        if version1 != version2: # not a voc
            raise SSException(vocfile.name + " does not appear to be a valid VOC file (version redundancy mismatch).")
        elif version1 != 0x010A and version1 != 0x0114: # not sure about the second one
            raise SSException("VOC version in " + vocfile.name + " is not supported by ScummVM.")
        
        # Go through and check if there's any unsupported blocks (ie not 1, 6, or 7)
        # Could write at the same time, but we want to verify that it's okay before commiting
        vocfile.seek(0)
        vocfile.seek(0x1A, 1) # skip rest of header (assume it's all fine) - block starts at 0x1A
        blocktype = ord(vocfile.read(1))
        while blocktype != 0:
            if blocktype != 1 and blocktype != 6 and blocktype != 7:
                raise SSException("Unsupported blocktype " + str(blocktype) + " in VOC file " + vocfile.name)
            size = strToNumber(vocfile.read(3)) # Read the block size (3 bytes)
            vocfile.seek(size, 1) # skip the block
            blocktype = ord(vocfile.read(1))
        
        vocfile.seek(0)
        
        self.tmpdata.flush()
        #self.tmpdata.write(ssconst.strVCTL)
        self.tmpdata.write("\x56\x43\x54\x4C" + numToString(8 + len(tags), 4))
        for t in tags: # Write the tags (gleaned from outside information)
            self.tmpdata.write(chr(t))
            self.tmpdata.flush()
        self.tmpdata.write(vocfile.read())
        self.tmpdata.flush()
        
        self.indextable.append( (start, [chr(t) for t in tags] ) )
        self.vocmap[start] = len(self.indextable) - 1
        self.descriptions.append(os.path.basename(vocfile.name)) # Store new file with new description
    
    def _addCompressedSound(self, infile, tags, infakepos=None):        
        # Read the input file (see what type it is)
        sHeader = infile.read(4)
        
        intype = determineSoundType(sHeader)
        
        if intype != self.type:
            raise SSException("Tried to add " + intype + " file when " + self.type + " was expected")
        
        infile.seek(0, 0)
            
        iNumTags = len(tags)
        
        # Append the converted data to the master output file
        self.tmpdata.flush()
        self.tmpdata.seek(0, 2) # seek to end for appending
        iOffset = self.tmpdata.tell()
        for t in tags: # Write the tags (gleaned from outside information)
            self.tmpdata.write(chr(t))
            self.tmpdata.flush()
        copyFileContents(infile, self.tmpdata) # copy OGG data
        self.tmpdata.flush()
        iTotSize = infile.tell() # size of OGG file
        
        # Metadata. If we're replacing a sound, don't append, do it in-place.
        if not infakepos is None:
            self.indextable[self.vocmap[infakepos]] = ([infakepos,
                            iOffset, # new offset
                            iNumTags, # num tags
                            iTotSize]) # compressed size
            # self.vocmap[self.fakepos] # don't adjust, it's already correct.
            self.descriptions[self.vocmap[infakepos]] = os.path.basename(infile.name)
        
        else:
            self.fakepos += 1 # fake original offset
            self.indextable.append([self.fakepos,
                            iOffset, # new offset
                            iNumTags, # num tags
                            iTotSize]) # compressed size
            self.vocmap[self.fakepos] = len(self.indextable) - 1
            self.descriptions.append(os.path.basename(infile.name))
        
        
        
    # End Adding sounds
    #______________
    # Removing sounds
    def removeSound(self, index):
        self.tmpdata.flush()
        self.tmpdata.seek(0)
        if self.type is "VOC":
            self._removeVOCSound(index)
        else:
            self._removeCompressedSound(index)
    
    
    def _removeVOCSound(self, index):
        
        # Actually remove the VOC from Monster file
        # Basically we're doing a list slice, getting file[:offset] + file[offset + self.tmpdata:]
        # Read all data before removed VOC
        offset = self.indextable[index][0]
        self.tmpdata.flush()
        #print offset
        predata = self.tmpdata.read(offset)
        self.tmpdata.flush()
        
        # If this isn't the last item in the monster.sou, keep everything after it
        if index < len(self.indextable) - 1:
            self.tmpdata.seek(self.indextable[index + 1][0], 0)
            postdata = self.tmpdata.read()
            self.tmpdata.flush()
        else:
            postdata = ""
        
        # Write concatenated data, then truncate
        self.tmpdata.seek(0)
        self.tmpdata.flush()
        self.tmpdata.write(predata + postdata)
        self.tmpdata.flush()
        self.tmpdata.truncate(self.tmpdata.tell())
        self.tmpdata.flush()
        

        # Reparse the text
        # If old index is found, the speech text will be removed
        # If index as per oldVocMap is > removed index, new offset is self.vocmap[index-1]
        oldVocMap = copy.copy(self.vocmap)
        remindex = oldVocMap[offset]
        
        self._createVOCTable(self.tmpdata) # Reload VOC mappings etc. Could be more efficient if we did it when we actually remove the VOC.
        
        # !
        # Rather than performing the below, we could make a new dict:
        # d[oldoffset] = newoffset
        # !
        
        for i, line in enumerate(ssglob.gTextData):
            if line[:8] == ssconst.strPrelude: # Has voice at the start of the string
                tmpstr1 = line[10:12] # LSB of VCTL address
                tmpstr2 = line[14:16]
                tmpstr3 = line[26:28]
                tmpstr4 = line[30:32] # MSB
                address = int((tmpstr4 + tmpstr3 + tmpstr2 + tmpstr1), 16)
                if address in oldVocMap:
                    if oldVocMap[address] == remindex: # this is the removed VOC so just remove the speech
                        ssglob.gTextData[i] = line[64:] # strip speech text
                    elif oldVocMap[address] > remindex: # the address is known
                        oldlineoffset = self.indextable[oldVocMap[address] - 1][0] # get the new address
                        offsethex = hex(oldlineoffset)[2:-1].zfill(8) # convert to hex
                        voxaddon = ssconst.strPrelude + \
                                   "\\x" + offsethex[6:8] + "\\x" + offsethex[4:6] + \
                                   ssconst.strPrelude + \
                                   "\\x" + offsethex[2:4] + "\\x" + offsethex[0:2] + \
                                   ssconst.strLastbit # WRONG but tags for VOC aren't implemented anyway
                        ssglob.gTextData[i] = voxaddon + line[64:] # replace the speech text
        
        # finally, remove the description
        #self.descriptions = self.descriptions[:remindex] + self.descriptions[remindex+1:]
        del self.descriptions[remindex]
        
    def _removeCompressedSound(self, index):       
        entry = self.indextable[index]
        origoffset = entry[0]
        actualoffset = entry[1]
        size = entry[3] + entry[2] # ogg data size + num of tags (each 1 byte)
        
        self.tmpdata.seek(0, 0)
        
        # This is a terrible way of doing it and will take up a jillion tons of memory
        self.tmpdata.flush()
        predata = self.tmpdata.read(actualoffset)
        self.tmpdata.flush()
        #self.tmpdata.seek(size, 1) # relative seeking doesn't seem to work for me anymore...
        
        # If this isn't the last item in the monster.sou, keep everything after it
        if index < len(self.indextable) - 1:
            self.tmpdata.seek(self.indextable[index + 1][1], 0)
            postdata = self.tmpdata.read()
            self.tmpdata.flush()
        else:
            postdata = ""
        
        # Removed sanity check for Indy 4/real tags
        #if postdata != "" and ord(postdata[0]) != MonsterSou.lDefaultTagList[0]:
        #    raise SSException("Data after removed sound did not match expected value.\n" + str(ord(postdata[0])))
        
        self.tmpdata.seek(0)
        self.tmpdata.flush()
        self.tmpdata.write(predata + postdata)
        self.tmpdata.flush()
        self.tmpdata.truncate(self.tmpdata.tell())
        self.tmpdata.flush()
        
        del predata
        del postdata


        # In every entry after removed index, adjust offsets and 
        #  vocmap links to indices.
        if index < len(self.indextable) - 1:
            for i in xrange((index + 1), len(self.indextable)):
                self.indextable[i][1] -= size
                self.vocmap[self.indextable[i][0]] -= 1
        
        
        #_______________
        # This should be moved to wherever removeSound is called from in the main App.
        #_______________
        # Remove speech text from any text referencing the removed sound.
        # Only need to fix the line(s) where the removed is referenced.
        # Don't need to fix lines that reference sounds after the removed one (as you do
        # for a non-compressed monster file), as the offsets used are fake and
        # arbitrary.
        for i, line in enumerate(ssglob.gTextData):
            if line[:8] == ssconst.strPrelude: # Has voice at the start of the string
                tmpstr1 = line[10:12] # LSB of VCTL address
                tmpstr2 = line[14:16]
                tmpstr3 = line[26:28]
                tmpstr4 = line[30:32] # MSB
                address = int((tmpstr4 + tmpstr3 + tmpstr2 + tmpstr1), 16)
                if address == origoffset:
                    # this is the removed VOC so just remove the speech
                    ssglob.gTextData[i] = line[64:] # strip speech text
        #_______________
        
        
        del self.vocmap[origoffset]
        del self.indextable[index]
        del self.descriptions[index]
        

    # End Removing sounds
    #______________
    def replaceSound(self, index, infile):
        self.tmpdata.flush()
        self.tmpdata.seek(0)
        if self.type is "VOC":
            raise SSException("Replace Sound is not supported for VOC-based MONSTER.SOU files.")
            #self._replaceVOCSound(index, infile)
        else:
            self._replaceCompressedSound(index, infile)

    def _replaceVOCSound(self, index, infile):
        """ Not implemented."""
        origoffset = self.indextable[index][0]

    def _replaceCompressedSound(self, index, infile):
        entry = self.indextable[index]
        origoffset = entry[0]
        actualoffset = entry[1]
        size = entry[3] + entry[2] # ogg data size + num of tags (each 1 byte)
        
        self.tmpdata.seek(0, 0)
        
        # This is a terrible way of doing it and will take up a jillion tons of memory
        # (in fact, we don't need to store the predata at all, we could just truncate
        #  up to the start of the file being removed. oh well!)
        self.tmpdata.flush()
        predata = self.tmpdata.read(actualoffset)
        self.tmpdata.flush()
        
        # Keep existing tags
        oldtags = self.tmpdata.read(entry[2])
        oldtags = [ ord(t) for t in oldtags ]
        
        # If this isn't the last item in the monster.sou, keep everything after it
        if index < len(self.indextable) - 1:
            self.tmpdata.seek(self.indextable[index + 1][1], 0)
            postdata = self.tmpdata.read()
            self.tmpdata.flush()
        else:
            postdata = ""
        
        # Truncate up to the removed sound, append the new sound, write the rest of the file.
        self.tmpdata.seek(0)
        self.tmpdata.flush()
        self.tmpdata.write(predata)
        self.tmpdata.flush()
        self.tmpdata.truncate(self.tmpdata.tell())
        self.tmpdata.flush()
        
        self._addCompressedSound(infile, tags=oldtags, infakepos=origoffset)
        self.tmpdata.flush()

        self.tmpdata.write(postdata)
        self.tmpdata.flush()
        self.tmpdata.truncate(self.tmpdata.tell())
        self.tmpdata.flush() # I really, really don't know when to flush.
        
        del predata
        del postdata

        # In every entry after removed index, adjust "actual" offsets.
        if index < len(self.indextable) - 1:
            entry = self.indextable[index]
            sizeofnew = entry[3] + entry[2] # could be nicer
            sizediff = size - sizeofnew
            for i in xrange((index + 1), len(self.indextable)):
                self.indextable[i][1] -= (size - sizeofnew)
        

    
    def exportSounds(self, outpath, speechmap):
        self.tmpdata.flush()
        self.tmpdata.seek(0)
        if self.type is "VOC":
            raise SSException("Export Sounds is not supported for VOC-based MONSTER.SOU files.")
        else:
            self._exportCompressedSounds(outpath, speechmap)
        
    def _exportVOCSounds(self, outpath, specehmap):
        for entry in self.indextable:
            origOffset = entry[0]
            rawTags = entry[1]
            numTags = len(rawTags)
            
             # Determine filename based on line number.
            # If sound is not used (no line number), use sound's description (may be a dummy).
            if origOffset in speechmap:
                desc = ssglob.gTextData[speechmap[origOffset][0]][64:]
                if len(desc) > 25: # limit filename length
                    desc = desc[:25]
                desc = ''.join(c for c in desc if c in ssconst.setValidChars)
                sfname = str(speechmap[origOffset][0]).zfill(5) + "_" + desc
                
                # For all but the first occurance of a sound that is used more than once,
                #  create "duplicate" marker files.
                if len(speechmap[origOffset]) > 1:
                    for ln in speechmap[origOffset][1:]:
                        dupname = str(ln).zfill(5) + "_" + desc
                        fname = os.path.join(outpath, dupname + '.dup')
                        if len(fname) > 255:
                            raise SSException("File name is too long: " + fname)
                        outfile = file(fname, 'wb')
                        outfile.write(str(speechmap[origOffset][0]))
                        outfile.close()
            
            else:
                desc = self.getDescriptionFromAddress(origOffset)
                if len(desc) > 20:
                    desc = desc[:20]
                sfname = "unk_" + str(origOffset).zfill(10) + "_" + \
                         ''.join(c for c in desc if c in ssconst.setValidChars)
            
            # Write the sound file.
            fname = os.path.join(outpath, sfname + '.' + self.type.lower())
            if len(fname) > 255:
                raise SSException("File name is too long: " + fname)
            outfile = file(fname, 'wb')
            self.tmpdata.flush()
            self.tmpdata.seek(origOffset + 8 + len(rawTags), 0)
            self.tmpdata.seek(self.indextable[index + 1][0])
            if index < len(self.indextable) - 1:
                outfile.write(self.tmpdata.read(self.indextable[0] - origOffset - numTags - 8))
            else:
                outfile.write(self.tmpdata.read())
            outfile.close()
            
            # Write the tags as well
            fname = os.path.join(outpath, sfname + ".tags")
            if len(fname) > 255:
                raise SSException("File name is too long: " + fname)
            outfile = file(fname, 'wb')
            outfile.write(rawTags)
            outfile.close()
        
        
    def _exportCompressedSounds(self, outpath, speechmap):
        for entry in self.indextable:
            origOffset = entry[0]
            actualOffset = entry[1]
            numTags = entry[2]
            soundSize = entry[3]
            
            # Determine filename based on line number.
            # If sound is not used (no line number), use sound's description (may be a dummy).
            if origOffset in speechmap:
                desc = ssglob.gTextData[speechmap[origOffset][0]][64:]
                if len(desc) > 25: # limit filename length
                    desc = desc[:25]
                desc = ''.join(c for c in desc if c in ssconst.setValidChars)
                sfname = str(speechmap[origOffset][0]).zfill(5) + "_" + desc
                
                # For all but the first occurance of a sound that is used more than once,
                #  create "duplicate" marker files.
                if len(speechmap[origOffset]) > 1:
                    for ln in speechmap[origOffset][1:]:
                        dupname = str(ln).zfill(5) + "_" + desc
                        fname = os.path.join(outpath, dupname + '.dup')
                        if len(fname) > 255:
                            raise SSException("File name is too long: " + fname)
                        outfile = file(fname, 'wb')
                        outfile.write(str(speechmap[origOffset][0]))
                        outfile.close()

            else:
                desc = self.getDescriptionFromAddress(origOffset)
                if len(desc) > 20:
                    desc = desc[:20]
                sfname = "unk_" + str(origOffset).zfill(10) + "_" + \
                         ''.join(c for c in desc if c in ssconst.setValidChars)
                         
            
            # Write the sound file.
            fname = os.path.join(outpath, sfname + '.' + self.type.lower())
            if len(fname) > 255:
                raise SSException("File name is too long: " + fname)
            outfile = file(fname, 'wb')
            self.tmpdata.flush()
            self.tmpdata.seek(actualOffset + numTags, 0)
            outfile.write(self.tmpdata.read(soundSize))
            outfile.close()

            # Write the tags as well
            fname = os.path.join(outpath, sfname + ".tags")
            if len(fname) > 255:
                raise SSException("File name is too long: " + fname)
            outfile = file(fname, 'wb')
            self.tmpdata.flush()
            self.tmpdata.seek(actualOffset, 0)
            outfile.write(self.tmpdata.read(numTags))
            outfile.close()